from typing import Optional, Type
from koil.task import KoilTask


class KoilState(object):
    def __init__(self, threaded=False, prefer_task=False, prefer_block=False) -> None:
        self.threaded = threaded
        self.prefer_task = prefer_task
        self.prefer_block = prefer_block
        super().__init__()

    def get_task_class(self) -> Optional[Type[KoilTask]]:
        return KoilTask



from typing import Dict, Type
from koil.checker.base import BaseChecker
from koil.state import KoilState
import logging

logger = logging.getLogger(__name__)


class CheckerRegistry:
    def __init__(self) -> None:
        self.checkerClasses: Dict[str, Type[BaseChecker]] = {}
        self.checkers: Dict[str, BaseChecker] = {}

    def register_checker(self, checker_class):
        self.checkerClasses[checker_class.__name__] = checker_class

    def get_checkers(self, koil):
        if not self.checkers:
            self.checkers = {
                key: checker_class(koil)
                for key, checker_class in self.checkerClasses.items()
            }
        return self.checkers

    def get_desired_state(self, koil):
        self.desired_states = {
            key: checker.force_state()
            for key, checker in self.get_checkers(koil).items()
        }
        self.desired_none_states = [
            state for key, state in self.desired_states.items() if state != None
        ]
        if len(self.desired_none_states) == 0:
            return KoilState(
                threaded=False
            )  # We are not in an event loop so lets stay sync motherfucker
        else:
            if not len(self.desired_none_states) == 1:
                logger.info(
                    f"More than one state returned from our checkers. This is not going to work {self.desired_none_states}"
                )
            return self.desired_none_states[0]


def register_checker(
    overwrite: bool = False,
    registry: CheckerRegistry = None,
):
    """Registers a Checker

    Registers a Structure with the a Package Registry. Once registered Arkitekt will use this Structure to expand and shrink Node requests.

    Args:
        identifier (str, optional): A unique identifier to be known be. Defaults to None.
        overwrite (bool, optional): If overwrite is set this Structure will replace exisiting Structures, otherwise we return an error on Reigstration
        registry (PackerRegistry, optional): Which registry to use in order to register this Strucute. Will use the default Registry if not set
    """

    def real_decorator(cls):
        (registry or get_checker_registry()).register_checker(cls)
        return cls

    return real_decorator


CHECKER_REGISTRY = None


def get_checker_registry(register_defaults=True):
    global CHECKER_REGISTRY
    if not CHECKER_REGISTRY:
        CHECKER_REGISTRY = CheckerRegistry()
        if register_defaults:
            import koil.checker.defaults

    return CHECKER_REGISTRY



from typing import Optional, Type, Union
from koil.checker.base import BaseChecker
from koil.checker.registry import register_checker
from koil.state import KoilState
from koil.task import KoilTask
import asyncio


try:
    from qtpy import QtWidgets
    from qtpy import QtCore

    class QtTask(KoilTask, QtCore.QObject):
        except_signal = QtCore.Signal(Exception)
        cancelled_signal = QtCore.Signal()
        done_signal = QtCore.Signal(object)

        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args, **kwargs)

        async def wrapped_future(self, future):
            try:
                value = await super().wrapped_future(future)
                self.done_signal.emit(value)
            except Exception as e:
                self.except_signal.emit(e)
            except asyncio.CancelledError:
                self.cancelled_signal.emit()

    HAS_QT = True

except Exception as e:
    QtTask = None
    QtWidgets = None
    HAS_QT = False


class QtKoilState(KoilState):
    def __init__(self, qt_app=None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.qt_app = qt_app

    def get_task_class(self) -> Optional[Type[KoilTask]]:
        return QtTask


@register_checker()
class QtChecker(BaseChecker):
    def force_state(self) -> Optional[KoilState]:
        """Checks if a running Qt Instance is there, if so we would like
        to run in a seperate thread

        Returns:
            Union[None, KoilState]: [description]
        """
        if HAS_QT:
            ap_instance = QtWidgets.QApplication.instance()
            if ap_instance is None:
                return None
            ap_instance.lastWindowClosed.connect(self.koil.close)

            return QtKoilState(qt_app=ap_instance, threaded=True, prefer_task=True)

        return None



import asyncio
from asyncio.futures import Future
from qtpy import QtGui
from qtpy.QtCore import QObject, Signal
import uuid
import logging
from koil.koil import Koil, get_current_koil

logger = logging.getLogger(__name__)


Reference = str


class UnconnectedSignalError(Exception):
    pass


def get_receiver_length(qobject, qsignal, callstring):
    try:
        return qobject.receivers(qsignal)
    except:
        return qobject.receivers(callstring)


class FutureWrapper(QObject):
    callSignal = Signal(Reference, tuple, dict)
    resolveSignal = Signal(Reference, object)
    rejectSignal = Signal(Reference, Exception)
    cancelSignal = Signal(Reference, tuple, dict)
    cancelledSignal = Signal(Reference)

    def wire(self, receiver, cancel_receiver=None):
        """Wires a Call signal to a QtPy Slot

        Args:
            receiver ([type]): [description]
            cancel_receiver ([type], optional): [description]. Defaults to None.
        """
        self.call_connected = True
        self.callSignal.connect(receiver)

        if cancel_receiver:
            self.cancel_connected = True
            self.cancelSignal.connect(cancel_receiver)

    def resolve(self, reference, value):
        self.resolveSignal.emit(reference, value)

    def reject(self, reference, value: Exception):
        self.rejectSignal.emit(reference, value)

    def cancelled(self, reference):
        self.cancelledSignal.emit(reference)

    def __init__(self, *args, cancel_timeout=4, pass_through=False, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.resolveSignal.connect(self.on_resolve)
        self.rejectSignal.connect(self.on_reject)
        self.cancelledSignal.connect(self.on_cancelled)
        self.futureMap = {}
        self.cancelMap = {}
        self.pass_through = pass_through
        self.cancel_timeout = 4

        self.call_connected = False
        self.cancel_connected = False

    def on_cancelled(self, ref):
        assert (
            ref in self.cancelMap
        ), "Received cancellation for non invoked cancel future"
        cancel_future = self.cancelMap[ref]
        if not cancel_future.done():
            self.loop.call_soon_threadsafe(cancel_future.set_result, None)

    def on_resolve(self, ref, obj):
        assert ref in self.futureMap, "Received resolvation for non invoked future"
        future = self.futureMap[ref]
        if not future.done():
            self.loop.call_soon_threadsafe(future.set_result, obj)

    def on_reject(self, ref, exp):
        assert ref in self.futureMap, "Received rejection for non invoked future"
        future = self.futureMap[ref]
        if not future.done():
            self.loop.call_soon_threadsafe(future.set_exception, exp)

    async def acall(self, *args, **kwargs):
        self.loop = asyncio.get_event_loop()
        if not self.call_connected:
            logger.info(f"Future {self} was never connected")
            if self.pass_through:
                return None

        reference = str(uuid.uuid4())
        future = self.loop.create_future()
        self.futureMap[reference] = future
        self.callSignal.emit(reference, args, kwargs)

        try:
            return await future
        except asyncio.CancelledError as e:
            if not self.cancel_connected:
                logger.info(f"Future {self} has no cancellation partner. Just Raising")
                raise e

            cancel_future = self.loop.create_future()
            self.cancelMap[reference] = future
            self.cancelSignal.emit(reference, args, kwargs)
            try:
                await asyncio.wait_for(cancel_future, timeout=self.cancel_timeout)
                raise e
            except asyncio.TimeoutError as t:
                logger.error("Cancellation recevied timeout, Cancelling anyways")
                raise e


class TimeoutFutureWrapper(FutureWrapper):
    def __init__(self, *args, timeout=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.timeout = timeout

    async def acall(self, *args, **kwargs):
        return await asyncio.wait_for(
            super().acall(*args, **kwargs), timeout=self.timeout
        )



import logging
import asyncio


logger = logging.getLogger(__name__)


class KoilTask:
    def __init__(
        self, future=None, koil=None, *args, log_errors=True, **kwargs
    ) -> None:
        super().__init__()
        self.future = future
        self.task = None
        self.log_errors = log_errors
        self.loop = koil.loop
        self.loop.call_soon_threadsafe(self.run)

    async def wrapped_future(self, future):
        try:
            return await future
        except Exception as e:
            if self.log_errors:
                logger.exception(e)
            raise e

    def run(self):
        self.task = self.loop.create_task(self.wrapped_future(self.future))

    async def acancel(self):
        try:
            self.task.cancel()
            try:
                await self.task
            except asyncio.CancelledError as e:
                logger.debug("Cancelled KoilTask")
        except Exception as e:
            logger.error("Koil Task Cancellation failed")

    def cancel(self):
        return asyncio.run_coroutine_threadsafe(self.acancel(), self.loop).result()



import contextvars
import threading
import asyncio
from asyncio.runners import _cancel_all_tasks
from threading import Thread
import os
import logging
import time

try:
    import uvloop
except:
    uvloop = None


from koil.checker.registry import get_checker_registry
from koil.state import KoilState

logger = logging.getLogger(__name__)


def newloop(loop, loop_started):
    asyncio.set_event_loop(loop)
    try:
        loop_started.set()
        logger.info("Running New Event loop in another Thread")
        loop.run_forever()
    finally:
        print("Loop Shutting Down")
        try:
            _cancel_all_tasks(loop)
            loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            asyncio.set_event_loop(None)
            loop.close()


class Koil:
    def __init__(
        self,
        force_sync=False,
        force_async=False,
        register_default_checkers=True,
        register_global=False,
        uvify=True,
        **overrides,
    ) -> None:
        """[summary]

        Args:
            force_sync (bool, optional): [description]. Defaults to False.
            force_async (bool, optional): [description]. Defaults to False.
            register_default_checkers (bool, optional): [description]. Defaults to True.
            register_global (bool, optional): [description]. Defaults to True.
            uvify (bool, optional): [description]. Defaults to True.
        """
        if uvify and uvloop is not None:
            uvloop.install()

        self.loop = None
        self.thread_id = None
        self.state = get_checker_registry(
            register_defaults=register_default_checkers
        ).get_desired_state(self)

        if force_sync or force_async:
            self.state.threaded = (
                force_sync and not force_async
            )  # Force async has priority

        if self.state.threaded:
            self.loop = asyncio.new_event_loop()
            self.loop_started_event = threading.Event()
            self.thread = Thread(
                name="Koil-Thread",
                target=newloop,
                args=(self.loop, self.loop_started_event),
            )
            self.thread.start()
            self.thread_id = self.thread.ident
            self.loop_started_event.wait()
            print(
                "Running in Thread. Make sure you close this loop. You can now use tasks!"
            )

        else:
            try:
                self.loop = asyncio.get_running_loop()
                self.thread_id = threading.current_thread().ident
                print(
                    "We are running in an event-loop. Better use the async api! We will do anyways"
                )
            except RuntimeError as e:
                self.loop = asyncio.new_event_loop()
                asyncio.set_event_loop(self.loop)
                self.thread_id = threading.current_thread().ident
                print("Created new eventloop in this. You can not use tasks!")

        if register_global:
            set_global_koil(self)

    async def aclose(self):
        loop = asyncio.get_event_loop()
        print("Causing loop to stop")
        loop.stop()
        print("Loop Stopped")

    def close(self):
        # Do according to state
        if self.state.threaded:
            print("Good you remembered to close the loop!")
            asyncio.run_coroutine_threadsafe(self.aclose(), self.loop)

            while self.loop.is_running():
                print("Waiting for the Loop to close")
                time.sleep(0.1)

    def __enter__(self):
        current_koil.set(self)
        return self

    def __exit__(self, *args, **kwargs):
        self.close()
        current_koil.set(None)


class KoiledContext:
    def __init__(self) -> None:
        pass

    def __enter__(self):
        self.koil = Koil(force_sync=True)
        current_koil.set(self.koil)
        return self.koil

    def __exit__(self, *args, **kwargs):
        self.koil.close()
        current_koil.set(None)
        return


current_koil = contextvars.ContextVar("current_koil", default=None)
GLOBAL_KOIL = None


koiled = KoiledContext()


def get_current_koil(**kwargs):
    global GLOBAL_KOIL
    koil = current_koil.get()
    if koil is None:
        if not GLOBAL_KOIL:
            GLOBAL_KOIL = Koil()
        return GLOBAL_KOIL

    return koil


def set_global_koil(koil):
    global GLOBAL_KOIL
    GLOBAL_KOIL = koil



import asyncio
import threading
import logging

from pkg_resources import load_entry_point


logger = logging.getLogger(__name__)


def unkoil(function):
    def koiled_function(*args, **kwargs):
        return koil(function(*args, **kwargs))

    return koiled_function


def koil(potential_future, as_task=False, koil=None, delay=False):
    """Taks an Async Function and according
    to the current loop setting either returns over
    it synchronusly in the herre loop or returns
    a future

    Args:
        potential_future ([type]): [description]

    Returns:
        [type]: [description]
    """
    from koil.koil import get_current_koil

    koil = koil or get_current_koil()
    state = koil.state
    loop = koil.loop

    if loop.is_running():
        # Loop is running
        if koil.thread_id == threading.current_thread().ident:
            # Loop is running in same thread, always returning future here (in an ansyc context you are the master)
            raise Exception(
                "You are calling an async function in the same thread as the loop that future is in, this is not allowed. Call the async function directly instead"
            )
        # We are in a threaded context our state is importat
        if as_task:
            return koil.state.get_task_class()(future=potential_future, koil=koil)

        if delay:
            return asyncio.run_coroutine_threadsafe(potential_future, loop)

        return asyncio.run_coroutine_threadsafe(potential_future, loop).result()

    # loop is not running and we have no threaded loop, just returning the futures
    assert (
        not as_task
    ), "It appears that you have a non threaded loop and you are trying to return a task, this is not possible. If this is happening try to initialize a running Koil in another thread with force_sync=True"

    print("Running Potential")
    return loop.run_until_complete(potential_future)


def next_on_gen(potential_generator, loop):
    """Takes a Async Generator and iterates over it
    threadsafe in the provided loop and returns the result
    synchronously in another generator

    Args:
        potential_generator ([type]): [description]
        loop ([type]): [description]

    Returns:
        [type]: [description]

    Yields:
        [type]: [description]
    """
    ait = potential_generator.__aiter__()

    async def next_on_ait():
        try:
            try:
                obj = await ait.__anext__()
                return False, obj
            except StopAsyncIteration:
                return True, None
        except asyncio.CancelledError as e:
            return True, Exception("Await cancelled Task")

    while True:
        done, obj = asyncio.run_coroutine_threadsafe(next_on_ait(), loop).result()
        if done:
            if obj:
                raise obj
            break
        yield obj


def next_on_gen_not_running(potential_generator, loop):
    """Takes a Async Generator and iterates over it
    threadsafe in the provided loop and returns the result
    synchronously in another generator

    Args:
        potential_generator ([type]): [description]
        loop ([type]): [description]

    Returns:
        [type]: [description]

    Yields:
        [type]: [description]
    """
    ait = potential_generator.__aiter__()

    async def next_on_ait():
        try:
            obj = await ait.__anext__()
            return False, obj
        except StopAsyncIteration:
            return True, None

    while True:
        done, obj = loop.run_until_complete(next_on_ait())
        if done:
            break
        yield obj


def koil_gen(potential_generator, koil=None):
    """Takes a Async Generator and iterates over it
    threadsafe in the herre loop providing a syncrhonus
    generator or returns an async generator in the current
    loop

    Args:
        potential_generator ([type]): [description]

    Returns:
        [type]: [description]
    """
    from koil.koil import get_current_koil

    koil = koil or get_current_koil()
    loop = koil.loop

    if loop.is_running():
        if koil.thread_id == threading.current_thread().ident:
            raise Exception(
                "You are calling an async generator in the same thread as the loop that future is in, this is not allowed. Call the async generator directly instead"
            )
        return next_on_gen(potential_generator, loop)

    return next_on_gen_not_running(potential_generator, loop)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['koil', 'koil.checker', 'koil.checker.defaults']

package_data = \
{'': ['*']}

extras_require = \
{'jupyter': ['jupyter>=1.0.0,<2.0.0']}

setup_kwargs = {
    'name': 'koil',
    'version': '0.1.66',
    'description': 'An abstract event loop to run in an async and sync context (especially jupyter and qt), with utility functions to allow both accesses',
    'long_description': '# koil\n\n[![codecov](https://codecov.io/gh/jhnnsrs/koil/branch/master/graph/badge.svg?token=UGXEA2THBV)](https://codecov.io/gh/jhnnsrs/koil)\n[![PyPI version](https://badge.fury.io/py/koil.svg)](https://pypi.org/project/koil/)\n![Maintainer](https://img.shields.io/badge/maintainer-jhnnsrs-blue)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/koil.svg)](https://pypi.python.org/pypi/koil/)\n[![PyPI status](https://img.shields.io/pypi/status/koil.svg)](https://pypi.python.org/pypi/koil/)\n[![PyPI download day](https://img.shields.io/pypi/dm/koil.svg)](https://pypi.python.org/pypi/koil/)\n\n### DEVELOPMENT\n\n### koil is asyncio for humans\n\nkoil is an abstraction layer on top of asyncio to enable "sensible defaults" for\nprogrammers working with frameworks that are barely compatible with asyncio (originally developped to get around pyqt5)\n\n### Why an abstraction layer on top of asyncio\n\nasyncio is a magificent addition to the python framework and is facilitating asyncronous\nand concurrent programming in a simple way.\n\nHowever especially in scientific and datascience contexts the benefits of diving\ninto the concepts of asyncio might not outweigh the costs of learning these paradigms.\n\nAdditionally there are huge players in the python ecoystem that do not work seemlessly\nwith asyncio yet and involve workarounds (Django, PyQt). Koil tries to encapsulate\nthese workarounds in a simple utility library that wraps your api in a safe context\nthat can be used form this frameworks.\n\n### Example\n\nAsyncio Way\n\n```python\nimport asyncio\n\nasync def call_api(sleep):\n    await asyncio.sleep(4)\n    return 5\n\n\nprint(asyncio.run(call_api(3)))\n\n```\n\nKoil Way\n\n```python\nimport asyncio\nfrom koil import koil\n\n@koil\nasync def call_api(sleep):\n    await asyncio.sleep(4)\n    return 5\n\nprint(call_api(sleep)) # note the ommision of the asyncio.run call\n\n```\n\nOn paper not a lot has changed between asyncio and koil, however under the hood the following\nhappened\n\nWhen calling call_api for the first time, koil detected that we are not running in an asyncronous\nevent loop and therefore wrapped the call in asyncio.run.\n\nThis doesn\'t justify a new library and a perfomance decrease though. So lets see this scenario\n\n```python\nimport asyncio\nfrom koil import koil\nfrom PyQt5.QtCore import QSize, Qt\nfrom PyQt5.QtWidgets import QApplication, QMainWindow, QPushButton\n\n@koil\nasync def call_api(interval):\n    await asyncio.sleep(interval)\n    return 5\n\n\nclass MainWindow(QMainWindow):\n    def __init__(self):\n        super().__init__()\n\n        self.setWindowTitle("My App")\n        button = QPushButton("Press Me!")\n        button.clicked.connect(self.button_clicked)\n        self.setCentralWidget(button)\n        self.sleep_task = None\n\n    def button_clicked(self):\n        if self.sleep_task: self.sleep_task.cancel()\n        self.sleep_task = call_api(interval, as_task=True)\n        self.sleep_task.resolve.connect(lambda x: self.button.setText(f"Slept {x} seconds"))\n\n\n\napp = QApplication(sys.argv)\n\nwindow = MainWindow()\nwindow.show()\n\napp.exec()\n\n```\n\nNow this is more koils problem set. Koil detected on first run (when creating the Koil object) that\nwe are running in an QtApplication instance and loaded our asyncornous task off to another thread\nto not interfer with our qt event loop. By passing as_task we also specified that we are do not\nwant to block the qt thread to wait for result in the other thread. Koil knowingly that we are in a Qt app\nallows us to now connect to signals of that specific task in the ui_thread.\n\nHere koil differnts itself of libraries like qasync that try to mimik the qt event loop in an asyncio event loop.\nWith qasync you adopt to the asyncio style and not the style you may want.\n\n### Jupyter notebook\n\nJupyter is in a unique position as it itself runs in an event loop but is mainly used by scientists\nthat are used to an syncronus interface.\n\nYour library\n\n```python\n\n@koil\nasync def call_api(interval):\n    await asyncio.sleep(interval)\n    return 5\n\n\n```\n\nBy default koil makes the assumption that if not specified differently it will use a synconrous\ninterface in juypter so.\n\n```python\n\ncall_api()\n\n```\n\nWould run your Api Task in another Thread and return the result synconrously back to you.\nHowever you can opt in to use a asyncronous api.\n\n```python\nfrom koil import Koil\n\nKoil(force_async=True) #at the top of your programm\n\n\nawait call_api()\n```\n\nMaking the call run in the same event loop of jupyter.\n\n### Considerations\n\nKoil is a library that doesnt want to exist. In general it provides convenience method\nto build asyncornous apis for a still synconrous world.\n\nTherefore when adding koil to your api make sure to not ditch your asyncrs api but extend\nthem through convenience methods. One such trick would be to leave your asyncornous functions\nuntouched and extend them with koiled versions.\n\n```python\nimport koil\n\nclass MyComplexApi:\n\n    def __init__(self):\n        ...\n\n    async def acall_endpoint(self, endpoint): #a as convention for asyncornous functions\n        ...\n\n\n    def call_endpoint(self, *args, *kwargs):\n        return koil(self.acall_endpoint)(*args, **kwargs)\n\n```\n\nLike this you are not experiencing any perfomance hints when running in a complete\nasyncornous context, but you enable unfamiliar users to use your application in a\nsyncronous world if they so desire.\n\nAlso it is wise to always call your pure async functions from other async functions.\n\n```python\nimport koil\nimport asyncio\n\nclass MyComplexApi:\n\n    def __init__(self):\n        self.connected = False\n        ...\n\n    async def aconnect(self):\n        await asyncio.sleep(3)\n        self.connected = True\n\n\n    async def acall_endpoint(self, endpoint):\n        if not self.connected:\n            await self.aconnect() # call asyncronous no koiled version for better performance\n\n        ...\n\n\n    def call_endpoint(self, *args, *kwargs):\n        return koil(self.acall_endpoint)(*args, **kwargs)\n\n```\n',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
